from __future__ import annotations

import hashlib
import os
import zipfile
from io import BytesIO
from pathlib import Path

directory_blacklist = {'.mypy_cache', '__pycache__', '.DS_Store'}
extension_blacklist = {'.py~'}
_bundle_path = '/wwwpy_bundle'


def _zip_path(zip_file, path):
    path = str(path)
    for root, dirs, files in os.walk(path):
        for file in files:
            zip_file.write(os.path.join(root, file),
                           os.path.relpath(os.path.join(root, file),
                                           os.path.join(path, '../remote')))


def zip_in_memory(path):
    stream = BytesIO()
    with zipfile.ZipFile(stream, "w", compression=zipfile.ZIP_DEFLATED, compresslevel=1) as zip_file:
        _zip_path(zip_file, path)

    stream.seek(0)
    return stream.getbuffer().tobytes()


def get_all_paths_with_hashes(path: str | Path) -> set[tuple[str, str | None]]:
    path = Path(path)
    paths = set()
    for path in path.rglob('*'):
        relative_path = path.relative_to(path)
        if path.is_file():
            file_hash = get_file_hash(path)
            paths.add((str(relative_path), file_hash))
        else:
            paths.add((str(relative_path), None))  # Directories don't need a hash
    return paths


def get_file_hash(filepath: str | Path) -> str:
    hash_sha256 = hashlib.sha256()
    with open(filepath, 'rb') as f:
        for chunk in iter(lambda: f.read(4096), b""):
            hash_sha256.update(chunk)
    return hash_sha256.hexdigest()


import gzip
import io
import base64


def bytes_gzip(content: bytes) -> bytes:
    buffer = io.BytesIO()
    with gzip.GzipFile(fileobj=buffer, mode='wb') as gz_file:
        gz_file.write(content)
    return buffer.getvalue()


def bytes_ungzip(compressed_data: bytes) -> bytes:
    buffer = io.BytesIO(compressed_data)
    with gzip.GzipFile(fileobj=buffer, mode='rb') as gz_file:
        return gz_file.read()


def str_gzip(content: str) -> bytes:
    return bytes_gzip(content.encode('utf-8'))


def str_ungzip(compressed_data: bytes) -> str:
    return bytes_ungzip(compressed_data).decode('utf-8')


def str_gzip_base64(content: str) -> str:
    """
    Compresses a string using gzip and encodes it with Base64.

    :param content: The input string to compress and encode.
    :return: A Base64-encoded string of the compressed data.
    """
    compressed_bytes = str_gzip(content)
    base64_encoded = base64.b64encode(compressed_bytes).decode('utf-8')
    return base64_encoded


def str_ungzip_base64(base64_string: str) -> str:
    """
    Decodes a Base64-encoded string and decompresses it using gzip.

    :param base64_string: The Base64-encoded compressed string.
    :return: The original uncompressed string.
    """
    compressed_bytes = base64.b64decode(base64_string)
    decompressed_str = str_ungzip(compressed_bytes)
    return decompressed_str



import os


async def app(scope, receive, send):
    if scope['type'] == 'http':
        await handle_http(scope, send)
    elif scope['type'] == 'websocket':
        await handle_websocket(scope, receive, send)


async def handle_http(scope, send):
    if scope['path'] == '/':
        await serve_file('index.html', 'text/html', send)
    else:
        await send_response(200, b'Hello, world!', 'text/plain', send)


async def handle_websocket(scope, receive, send):
    if scope['path'] != '/echo':
        return
    await send({'type': 'websocket.accept'})

    while True:
        message = await receive()
        if message['type'] == 'websocket.receive':
            text = message.get('text')
            await send({'type': 'websocket.send', 'text': f"echo -> {text}"})
        elif message['type'] == 'websocket.disconnect':
            break


async def serve_file(filename, content_type, send):
    try:
        with open(filename, "rb") as f:
            content = f.read()
        await send_response(200, content, content_type, send)
    except FileNotFoundError:
        await send_response(404, b'File not found', 'text/plain', send)


async def send_response(status, body, content_type, send):
    await send({
        'type': 'http.response.start',
        'status': status,
        'headers': [[b'content-type', content_type.encode()], ],
    })
    await send({'type': 'http.response.body', 'body': body, })



from __future__ import annotations

from inspect import iscoroutinefunction
import asyncio
import js
from js import console

import wwwpy.common.reloader as reloader
from wwwpy.common import _no_remote_infrastructure_found_text
from wwwpy.common.tree import print_tree
from wwwpy.remote.designer import dev_mode as dm
from wwwpy.remote.websocket import setup_websocket


async def entry_point(dev_mode: bool = False):
    # from wwwpy.common.tree import print_tree
    # print_tree('/wwwpy_bundle')

    await setup_websocket()
    if dev_mode:
        await dm.activate()

    await _invoke_browser_main()


def _reload():
    async def reload():
        import wwwpy.common.modlib as modlib
        console.log('reloading')
        for package in ['remote', 'common']:
            directory = modlib._find_package_directory(package)
            if directory:
                reloader.unload_path(str(directory))
        await _invoke_browser_main(True)

    asyncio.create_task(reload())


async def _invoke_browser_main(reload=False):
    try:
        console.log('invoke_browser_main')
        if dm.is_active():
            from wwwpy.remote.designer import helpers
            js.window.onerror = helpers._on_error
            js.window.onunhandledrejection = helpers._on_unhandledrejection
            from wwwpy.remote.designer import log_redirect
            log_redirect.redirect_logging()

        try:
            js.document.body.innerText = f'Going to import the "remote" package'
            import remote
            if reload:
                import importlib
                importlib.reload(remote)
        except ImportError as e:
            import traceback
            msg = _no_remote_infrastructure_found_text + ' Exception: ' + str(
                e) + '\n\n' + traceback.format_exc() + '\n\n'
            js.document.body.innerHTML = msg.replace('\n', '<br>')
            return

        if hasattr(remote, 'main'):
            if iscoroutinefunction(remote.main):
                await remote.main()
            else:
                remote.main()
    finally:
        if dm.is_active():
            from wwwpy.remote.designer.ui import dev_mode_component
            dev_mode_component.show()



import js
import wwwpy.remote.component as wpc

class RpcIssueAlert(wpc.Component):
    button1: js.HTMLButtonElement = wpc.element()

    def init_component(self):
        # language=html
        self.element.innerHTML = """
<h2>Server.rpc issue:
The package server.rpc is not completely supported by the hot-reload yet.
It is necessary to refresh the browser. <br>
<a href='https://github.com/www-py/wwwpy/issues/6' target='_blank'>See and vote the github issue</a>
<br>
<button data-name="button1" style='transform: scale(2.0,2.0); margin: 2em 4em'>Refresh page</button>
</h2>

        """
        self.element.style.position = 'absolute'
        self.element.style.top = '0'
        self.element.style.display = 'none'
        self.element.style.backgroundColor = 'red'
        try:
            from server import rpc
        except ImportError as e:
            self.element.style.display = 'block'

    async def button1__click(self, event):
        js.console.log('handler button1__click event =', event)
        js.location.reload()
    



from __future__ import annotations

import dataclasses
from dataclasses import dataclass
from optparse import Option
from pathlib import Path
from typing import AnyStr, Union, Optional


@dataclass(frozen=True)
class Event:
    event_type: str
    is_directory: bool
    src_path: str
    dest_path: str = ''
    content: Union[str, bytes, None] = None

    def strip_container(self, container_path: str) -> 'Event':
        if (not self.src_path.startswith(container_path) or (
                self.dest_path and not self.dest_path.startswith(container_path))):
            raise ValueError(f'Paths are not contained in {container_path} self={self}')

        len_cont = len(container_path)

        def fix(path: str) -> str:
            if path == '':
                return ''
            new_path = path[len_cont:]
            if new_path == '':
                return '/'
            return new_path

        src_path = fix(self.src_path)
        dest_path = fix(self.dest_path)

        return Event(self.event_type, self.is_directory, src_path, dest_path)

    def to_absolute(self, into: Path) -> 'Event':
        e = self
        src_path = (into / e.src_path).as_posix()
        dest_path = '' if e.dest_path == '' else (into / e.dest_path).as_posix()
        return dataclasses.replace(e, src_path=src_path, dest_path=dest_path)

    def relative_to(self, other):
        e = self
        if e.src_path == '' or e.src_path == '/':
            raise ValueError(f'Invalid src_path: {e.src_path}')
        src = Path(e.src_path).relative_to(other)
        res = dataclasses.replace(e, src_path=src.as_posix())

        if e.dest_path == '':
            return res
        if e.dest_path and (e.dest_path == '' or e.dest_path == '/'):
            raise ValueError(f'Invalid dest_path: {e.dest_path}')
        dst = Path(e.dest_path).relative_to(other)
        res2 = dataclasses.replace(res, dest_path=dst.as_posix())
        return res2



from __future__ import annotations

from typing import Callable

from wwwpy.common.rpc.serializer import RpcRequest
import asyncio

import logging

logger = logging.getLogger(__name__)


async def setup_websocket():
    from js import window, console
    import importlib
    def log(msg):
        console.log(msg)

    def message(msg):
        log(f'message:{msg}')
        r = RpcRequest.from_json(msg)
        # _debug_requested_module(r.module)
        m = importlib.import_module(r.module)
        class_name, func_name = r.func.split('.')
        attr = getattr(m, class_name)
        inst = attr()
        func = getattr(inst, func_name)
        func(*r.args)

    l = window.location
    proto = 'ws' if l.protocol == 'http:' else 'wss'
    url = f'{proto}://{l.host}/wwwpy/ws'
    _WebSocketReconnect(url, message)


class _WebSocketReconnect:
    def __init__(self, url: str, on_message: Callable):
        self._url = url
        self._on_message = on_message
        self._counter = 0
        self._connect()

    def _connect(self):
        from js import WebSocket, console
        self._counter += 1
        console.log(f'connecting to {self._url} counter={self._counter}')
        es = WebSocket.new(self._url)
        es.onopen = lambda e: console.log('open')
        es.onmessage = lambda e: self._on_message(e.data)
        es.onerror = lambda e: es.close()

        async def reconnect():
            await asyncio.sleep(1)
            self._connect()

        es.onclose = lambda e: asyncio.create_task(reconnect())  # for now, we just retry forever and ever


def _debug_requested_module(module_name: str):
    from wwwpy.common import modlib
    mp = modlib._find_module_path(module_name)
    logger.debug(f'requested module: {module_name} path: {mp}')
    if mp:
        logger.debug(f'content: ```{mp.read_text()}```')



from __future__ import annotations

from dataclasses import dataclass
from pathlib import Path

from js import console

from wwwpy.common import modlib


@dataclass
class _Dir:
    root: Path
    remote: Path


def _get_dir():
    # this works because both wwwpy and user code are flattened in the same folder
    root = modlib._find_module_path('wwwpy').parent.parent
    console.log(f'root={root}')
    remote = root / 'remote'
    return _Dir(root, remote)



from __future__ import annotations

import base64
import os
import zipfile
from io import BytesIO
from pathlib import Path

import shutil
from pathlib import Path
from typing import List, Any

from wwwpy.common.filesystem.sync import Event


def sync_source(source: Path, events: List[Event]) -> List[Any]:
    if not events:
        return []
    return sync_init(source)


def _delete_recursive(path: Path, ignore_errors):
    for e in path.iterdir():
        if e.is_file():
            e.unlink(missing_ok=ignore_errors)
        elif e.is_dir():
            shutil.rmtree(e, ignore_errors=ignore_errors)


def sync_target(target_root: Path, changes: List[Any]) -> None:
    if not changes:
        return
    _delete_recursive(target_root, ignore_errors=True)
    _delete_recursive(target_root, ignore_errors=False)

    b = base64.b64decode(changes[0])
    with BytesIO(b) as stream:
        with zipfile.ZipFile(stream, "r") as zip_file:
            zip_file.extractall(target_root)


def sync_init(source: Path) -> List[Any]:
    b = _zip_in_memory(source)
    s = base64.b64encode(b).decode('utf-8')
    return [s]


def _zip_path(zip_file, path):
    if os.path.isfile(path):
        zip_file.write(path, os.path.basename(path))
    elif os.path.isdir(path):
        for root, dirs, files in os.walk(path):
            for file in files:
                file_path = os.path.join(root, file)
                arcname = os.path.relpath(file_path, path)
                zip_file.write(file_path, arcname)


def _zip_in_memory(path) -> bytes:
    stream = BytesIO()
    with zipfile.ZipFile(stream, "w", compression=zipfile.ZIP_DEFLATED, compresslevel=1) as zip_file:
        _zip_path(zip_file, path)

    stream.seek(0)
    return stream.getbuffer().tobytes()



class WwwpyException(Exception):
    pass


class RemoteException(WwwpyException):
    """When the server code fails"""
    pass


class RemoteError(WwwpyException):
    """When network/wwwpy infrastructure fails"""
    pass



import subprocess

# pip install tornado

import tornado.ioloop
import tornado.web
import tornado_asgi_handler
import echo_handler

application = tornado.web.Application([
    (r".*", tornado_asgi_handler.AsgiHandler, dict(asgi_app=echo_handler.app))
])
application.listen(8000)

try:
    tornado.ioloop.IOLoop.current().start()
except KeyboardInterrupt as err:
    print("Server stopped")



from __future__ import annotations

import sys
import time
from pathlib import Path

from wwwpy.bootstrap import bootstrap_routes
from wwwpy.server.custom_str import CustomStr
from wwwpy.common.designer import log_emit
from wwwpy.common.rpc.custom_loader import CustomFinder
from wwwpy.resources import library_resources, from_directory, from_file
from wwwpy.webserver import Webserver
from wwwpy.webservers.available_webservers import available_webservers
from wwwpy.websocket import WebsocketPool
from wwwpy.server import tcp_port

import logging

logger = logging.getLogger(__name__)


def start_default(port: int, directory: Path, dev_mode=False):
    webserver = available_webservers().new_instance()

    convention(directory, webserver, dev_mode=dev_mode)

    while tcp_port.is_port_busy(port):
        logger.warning(f'port {port} is busy, retrying...')
        [time.sleep(0.1) for _ in range(20) if tcp_port.is_port_busy(port)]

    webserver.set_port(port).start_listen()


websocket_pool: WebsocketPool = None


def convention(directory: Path, webserver: Webserver = None, dev_mode=False):
    print(f'applying convention to working_dir: {directory}')
    server_rpc_packages = ['server.rpc']
    remote_rpc_packages = {'remote', 'remote.rpc', 'wwwpy.remote', 'wwwpy.remote.rpc'}
    if dev_mode:
        server_rpc_packages.append('wwwpy.server.designer.rpc')
        remote_rpc_packages.add('wwwpy.remote.designer')
        remote_rpc_packages.add('wwwpy.remote.designer.rpc')
        log_emit.add_once(print)
        from wwwpy.common import quickstart
        quickstart._make_hotreload_work(directory)

    sys.path.insert(0, CustomStr(directory))
    sys.meta_path.insert(0, CustomFinder(remote_rpc_packages))
    global websocket_pool
    websocket_pool = WebsocketPool('/wwwpy/ws')
    services = _configure_server_rpc_services('/wwwpy/rpc', server_rpc_packages)
    routes = [services.route, websocket_pool.http_route, *bootstrap_routes(
        resources=[
            library_resources(),
            services.remote_stub_resources(),
            from_directory(directory / 'remote', relative_to=directory),
            from_directory(directory / 'common', relative_to=directory),
            from_file(directory / 'common.py', relative_to=directory),  # remove .py support
            from_file(directory / 'remote.py', relative_to=directory),  # remove .py support
        ],
        # language=python
        python=f'from wwwpy.remote.browser_main import entry_point; await entry_point(dev_mode={dev_mode})'
    )]

    if dev_mode:
        from wwwpy.server.designer import dev_mode as dm
        dm._hotreload_remote(['common', 'remote'], websocket_pool)
        dm._hotreload_server(['common', 'server'])
        dm._warning_on_multiple_clients(websocket_pool)


    if webserver is not None:
        webserver.set_http_route(*routes)


from wwwpy.rpc import RpcRoute


def _configure_server_rpc_services(route_path: str, modules: list[str]) -> RpcRoute:
    services = RpcRoute(route_path)

    for module_name in modules:
        try:
            services.add_module(module_name)
        except Exception as e:
            print(f'could not load rpc module `{module_name}`: {e}')
            return None

    return services



import inspect
import os
from threading import Thread
from types import FunctionType
from typing import Callable, Any

from playwright.sync_api import PageAssertions, LocatorAssertions, APIResponseAssertions, Page, sync_playwright, \
    Playwright

with sync_playwright() as pw: pass  # workaround to run playwright in a new thread. see: https://github.com/microsoft/playwright-python/issues/1685


def start_playwright_in_thread(url: str, headless: bool):
    def in_thread():
        start_playwright(url, headless)

    thread = Thread(target=in_thread, daemon=True)
    thread.start()


def start_playwright(url: str, headless: bool):
    playwright = sync_playwright().start()
    browser = playwright.chromium.launch(headless=headless)
    page = browser.new_page()
    playwright_setup_page_logger(page)
    page.goto(url)
    return playwright


def playwright_setup_page_logger(page: Page):
    page.on('console', lambda msg: print(f'console [{msg.type}] ==== {msg.text}'))
    sep = '\n' + ('=' * 60) + '\n'
    page.on('pageerror', lambda exc: print(f'{sep}uncaught exception through pageerror: {sep}{exc}{sep}'))


def playwright_patch_timeout() -> None:
    def PLAYWRIGHT_PATCH_TIMEOUT_MILLIS() -> int:
        timeout = 30000
        try:
            import wwwpy_user_conf
            timeout = wwwpy_user_conf.PLAYWRIGHT_PATCH_TIMEOUT_MILLIS
        except:
            pass
        return int(os.environ.get('PLAYWRIGHT_PATCH_TIMEOUT_MILLIS', f'{timeout}'))

    print(f'Using PLAYWRIGHT_PATCH_TIMEOUT, current value={PLAYWRIGHT_PATCH_TIMEOUT_MILLIS()}')

    # patch playwright assertion timeout to match our configuration
    # this is temporary solution until playwright supports setting custom timeout for assertions
    # github issue: https://github.com/microsoft/playwright-python/issues/1358

    def patch_timeout(_member_obj: FunctionType) -> Callable:
        def patch_timeout_inner(*args, **kwargs) -> Any:
            __tracebackhide__ = True
            timeout_millis = PLAYWRIGHT_PATCH_TIMEOUT_MILLIS()
            parameters = inspect.signature(_member_obj).parameters
            timeout_arg_index = list(parameters.keys()).index("timeout")
            if timeout_arg_index >= 0:
                if len(args) > timeout_arg_index:
                    args = list(args)  # type: ignore
                    args[timeout_arg_index] = timeout_millis  # type: ignore
                elif 'timeout' not in kwargs:
                    kwargs["timeout"] = timeout_millis
            return _member_obj(*args, **kwargs)

        return patch_timeout_inner

    for assertion_cls in [PageAssertions, LocatorAssertions, APIResponseAssertions]:
        for member_name, member_obj in inspect.getmembers(assertion_cls):
            if isinstance(member_obj, FunctionType):
                if "timeout" in inspect.signature(member_obj).parameters:
                    setattr(assertion_cls, member_name, patch_timeout(member_obj))



from __future__ import annotations

import inspect
import zipfile
from abc import ABC, abstractmethod
from dataclasses import dataclass
from io import BytesIO
from pathlib import Path
from typing import Iterator, Callable, Optional, TypeVar, Iterable, Protocol, Tuple
from zipfile import ZipFile

from wwwpy.common import iterlib, modlib, files
from wwwpy.common.iterlib import CallableToIterable

parent = Path(__file__).resolve().parent


@dataclass(frozen=True)
class Resource(ABC):
    """represents something (a file) to be transferred to the remote"""
    arcname: str

    @abstractmethod
    def _bytes(self) -> bytes:
        pass


ResourceIterable = Iterable[Resource]
"""Let's say that the Iterable is a factory function that returns an Iterator.
https://docs.python.org/3/library/collections.abc.html<
https://wiki.python.org/moin/Iterator
"""


@dataclass(frozen=True)
class StringResource(Resource):
    content: str

    def _bytes(self) -> bytes:
        return self.content.encode('utf-8')


@dataclass(frozen=True)
class PathResource(Resource):
    filepath: Path

    def _bytes(self) -> bytes:
        raise Exception('this has a special method to get the bytes')


TResource = TypeVar("TResource", bound=Resource)

ResourceAccept = Callable[[TResource], bool]


def default_resource_accept(resource: Resource) -> bool:
    if not isinstance(resource, PathResource):
        return True
    filepath = resource.filepath
    if filepath.name == '.DS_Store':
        return False
    if filepath.is_dir() and filepath.name in files.directory_blacklist:
        return False
    return True


FilesystemIterable = Iterable[PathResource]


def from_file(file: Path, relative_to: Path) -> FilesystemIterable:
    def bundle() -> Iterator[PathResource]:
        rel = file.relative_to(relative_to)
        if file.is_file() and file.exists():
            yield PathResource(str(rel), file)

    return CallableToIterable(bundle)


def _recurse(path: Path, relative_to: Path, resource_accept: ResourceAccept) -> Iterator[PathResource]:
    for f in path.glob('*'):
        rel = f.relative_to(relative_to)
        candidate = PathResource(str(rel), f)
        if resource_accept(candidate):
            if f.is_file():
                yield candidate
            else:
                yield from _recurse(f, relative_to, resource_accept)


def from_directory(
        folder: Path, relative_to: Path | None = None,
        resource_accept: ResourceAccept = default_resource_accept
) -> FilesystemIterable:
    return from_directory_lazy(lambda: (folder, relative_to), resource_accept)


class FolderProvider(Protocol):
    """Return tuple of two Paths[folder, relative_to]. Both are optional"""

    def __call__(self) -> Tuple[Path | None, Path | None]: ...


def from_directory_lazy(
        folder_provider: FolderProvider,
        resource_accept: ResourceAccept = default_resource_accept
) -> FilesystemIterable:
    def bundle() -> Iterator[PathResource]:
        folder, relative_to = folder_provider()
        if folder is None:
            return
        rel_to: Path = folder if relative_to is None else relative_to
        yield from _recurse(folder, rel_to, resource_accept)

    return CallableToIterable(bundle)


def build_archive(resource_iterator: Iterator[Resource]) -> bytes:
    """builds a zip archive from the given resources and returns the bytes"""

    stream = BytesIO()
    zip_file = ZipFile(stream, "w", compression=zipfile.ZIP_DEFLATED, compresslevel=1)
    for resource in iterlib.iter_catching(resource_iterator):
        if isinstance(resource, PathResource):
            zip_file.write(resource.filepath, resource.arcname)
        elif isinstance(resource, Resource):
            zip_file.writestr(resource.arcname, resource._bytes())
        else:
            raise Exception(f'Unhandled class \n  type={type(resource).__name__} \n  data={resource}')
    zip_file.close()

    stream.seek(0)
    return stream.getbuffer().tobytes()


def stacktrace_pathfinder(stack_backtrack: int = 1) -> Optional[Path]:
    wwwpy_root = parent

    for stack in inspect.stack():
        source_path = Path(stack.filename).resolve()
        if not _is_path_contained(source_path, wwwpy_root):
            stack_backtrack -= 1
            if stack_backtrack == 0:
                return source_path

    return None


def _is_path_contained(child: Path, parent: Path) -> bool:
    cl = len(child.parts)
    cp = len(parent.parts)
    if cl < cp:
        return False
    m = min(cl, cp)
    child_parts = child.parts[:m]
    parent_parts = parent.parts[:m]
    return child_parts == parent_parts


def library_resources() -> Iterable[PathResource]:
    """returns the resources from the wwwpy library itself"""

    # return from_directory(parent, relative_to=parent.parent)
    def accept(resource: Resource) -> bool:
        if not default_resource_accept(resource):
            return False
        path = Path(resource.arcname)
        if path.parts[0] != 'wwwpy':
            return False
        if len(path.parts) == 2:
            return True  # wwwpy/* root files
        return path.parts[1] in {'common', 'remote'}

    return from_directory_lazy(lambda: (parent, parent.parent), accept)


def _fs_iterable(package_name: str) -> Tuple[Path | None, Path | None]:
    target = modlib._find_package_directory(package_name)
    if not target:
        return None, None
    root = target.parent
    r = range(package_name.count('.'))
    for _ in r:
        root = root.parent

    return target, root


def from_module_lazy(module_name: str):
    return from_directory_lazy(lambda: _fs_iterable(module_name))



from __future__ import annotations

import json
import threading
from functools import partial
from pathlib import Path
from queue import Queue
from typing import Tuple

from xvirt import XVirt

from wwwpy.bootstrap import bootstrap_routes
from wwwpy.common import modlib
from wwwpy.http import HttpRoute, HttpRequest, HttpResponse
from wwwpy.resources import library_resources, StringResource, from_directory_lazy
from wwwpy.server.tcp_port import find_port
from .playwrightlib import start_playwright_in_thread
from wwwpy.webservers.available_webservers import available_webservers

_file_parent = Path(__file__).parent


class XVirtImpl(XVirt):

    def __init__(self, headless: bool = True):
        self.headless = headless

        self.events = Queue()
        self.close_pw = threading.Event()

    def virtual_path(self) -> str:
        location = modlib._find_module_path('tests.remote')
        if not location:
            return 'tests.remote-not-available'
        return str(location.parent)

    def _http_handler(self, req: HttpRequest) -> HttpResponse:
        print(f'server side xvirt_notify_handler({req})')
        self.events.put(req.content)
        return HttpResponse('', 'text/plain')

    def run(self):
        webserver = self._start_webserver()
        start_playwright_in_thread(webserver.localhost_url(), self.headless)

    def _start_webserver(self):
        xvirt_notify_route = HttpRoute('/xvirt_notify', self._http_handler)
        # read remote conftest content
        remote_conftest = (_file_parent / 'remote_conftest.py').read_text() \
            .replace('#xvirt_notify_path_marker#', '/xvirt_notify')

        def fs_iterable(package_name: str) -> Tuple[Path | None, Path | None]:
            target = modlib._find_package_directory(package_name)
            if not target:
                return None, None
            root = target.parent
            r = range(package_name.count('.'))
            for _ in r:
                root = root.parent

            return target, root

        resources = [library_resources(),
                     from_directory_lazy(partial(fs_iterable, 'remote')),
                     from_directory_lazy(partial(fs_iterable, 'common')),
                     from_directory_lazy(partial(fs_iterable, 'tests.remote')),
                     from_directory_lazy(partial(fs_iterable, 'tests.common')),
                     [
                         StringResource('tests/__init__.py', ''),
                         StringResource('pytest.ini', ''),
                         StringResource('conftest.py', remote_conftest),
                         StringResource('remote_test_main.py',
                                        (_file_parent / 'remote_test_main.py').read_text())],

                     ]
        webserver = available_webservers().new_instance()
        invocation_dir, args = self.remote_invocation_params('/wwwpy_bundle')
        invocation_dir_json = json.dumps(invocation_dir)
        args_json = json.dumps(args)
        rootpath = json.dumps('/wwwpy_bundle')
        bootstrap_python = f'import remote_test_main; await remote_test_main.main({rootpath},{invocation_dir_json},{args_json})'
        webserver.set_http_route(*bootstrap_routes(resources, python=bootstrap_python), xvirt_notify_route)
        webserver.set_port(find_port()).start_listen()
        return webserver

    def finalize(self):
        self.close_pw.set()
        # self._thread.join()

    def recv_event(self) -> str:
        return self.events.get(timeout=30)



# Browse pyscript virtual filesystem
from __future__ import annotations

import os
from pathlib import Path
from typing import List

import js
from js import console

from wwwpy.common.files import zip_in_memory
from wwwpy.remote.designer.ui.draggable_component import DraggableComponent, new_window
from wwwpy.remote.files import download_bytes
import wwwpy.remote.component as wpc


class FilesystemTree(wpc.Component):
    path: Path
    _indent: int = 0
    children: List[FilesystemTree]

    _entity: js.HTMLElement = wpc.element()
    entity_actions: js.HTMLElement = wpc.element()
    _children: js.HTMLElement = wpc.element()
    _download: js.HTMLElement = wpc.element()
    _view: js.HTMLElement = wpc.element()

    def init_component(self):
        # language=html
        self.element.innerHTML = """<div>
    <span data-name="_entity" style="cursor: pointer"></span>
    <span data-name="entity_actions" style="cursor: pointer">
        <span data-name="_download" >&nbsp ⬇</span>
        <span data-name="_view"> 👁 &nbsp</span>
    </span>
    <div data-name="_children" style="margin-left: 1em"></div>
</div>
        """

    async def _entity__click(self, event):
        self._toggle_display()

    async def _download__click(self, event):
        self._do_download()

    def show_path(self, path: str | Path, indent=1):
        self.path = Path(path)
        self._indent = indent
        self.children = []
        self._view.style.display = 'none' if self.is_dir() else ''

        if self._indent > 1:
            self._toggle_display()

        self._update_caption()

        if self.is_dir():
            self._recurse()

    def is_dir(self):
        # return self.path.is_dir() # in pyodide: `PermissionError: [Errno 63] Operation not permitted: '/proc/self/fd'`
        return os.path.isdir(self.path)

    def is_file(self):
        return os.path.isfile(self.path)

    def _toggle_display(self, *args):
        self._children.style.display = '' if self._children_hidden() else 'none'
        self._update_caption()

    def _do_download(self, *args):
        console.log(f'click {self.path.absolute()}')
        if self.is_dir():
            download_bytes(self.path.name + '.zip', zip_in_memory(self.path))
        else:
            download_bytes(self.path.name, self.path.read_bytes())

    def _update_caption(self):
        if_dir = '▸' if self._children_hidden() else '▼'
        pre = if_dir if self.is_dir() else ' '
        self._entity.innerHTML = (f'<span style="display:inline-block; width: 1em">{pre}</span>'
                                  + ' ' + self.path.name)

    def _children_hidden(self):
        s = self._children.style
        return not s.display == ''

    def _recurse(self):
        for child in self.path.glob('*'):
            w = FilesystemTree()
            w.show_path(child, self._indent + 1)
            self.children.append(w)
            self._children.append(w.root_element())

    def _view__click(self, event):
        if not self.is_file():
            return
        win = new_window(f'{self.path}', closable=True)
        component = win.window
        js.document.body.append(component.element)

        pre1: js.HTMLElement = js.document.createElement('pre')
        pre1.innerText = self.path.read_text()
        win.element.append(pre1)


def show_explorer():
    import js

    component = new_window('Browse local filesystem')
    tree = FilesystemTree()
    tree.show_path('/')
    component.element.append(tree.element)
    # todo use DevModeComponent.instance
    js.document.body.append(component.element)



import asyncio
import base64
import logging

import js
import wwwpy.remote.component as wpc
from pyodide.ffi import create_proxy, JsException

logger = logging.getLogger(__name__)


class UploadComponent(wpc.Component, tag_name='wwwpy-quickstart-upload'):
    file_input: js.HTMLInputElement = wpc.element()
    uploads: js.HTMLElement = wpc.element()

    @property
    def multiple(self) -> bool:
        return self.file_input.hasAttribute('multiple')

    @multiple.setter
    def multiple(self, value: bool):
        self.file_input.toggleAttribute('multiple', value)

    def init_component(self):
        # language=html
        self.element.innerHTML = """
<input data-name="file_input" placeholder="input1" type="file" multiple>
<div data-name="uploads"></div>
        """

    async def file_input__change(self, event):
        logger.info('file_input__change')
        self.uploads.innerHTML = ''
        self.file_input.style.display = 'none'
        files = self.file_input.files
        tasks = []
        for file in files:
            progress = UploadProgressComponent()
            self.uploads.appendChild(progress.element)
            tasks.append(progress.upload(file))
        await asyncio.gather(*tasks)
        self.uploads.insertAdjacentHTML('beforeend', 'Upload complete')
        await asyncio.sleep(3)
        self.file_input.style.display = 'block'
        self.file_input.value = ''
        self.uploads.innerHTML = ''


class UploadProgressComponent(wpc.Component):
    file_input: js.HTMLInputElement = wpc.element()
    progress: js.HTMLProgressElement = wpc.element()
    progress_label: js.HTMLLabelElement = wpc.element()
    label: js.HTMLInputElement = wpc.element()

    def init_component(self):
        # language=html
        self.element.innerHTML = """
<div data-name="progress_label" style='padding: 0.5em'>
    <progress data-name="progress"></progress>
    &nbsp;<span data-name="label"></span>
</div>"""

    async def upload(self, file: js.File, chunk_size: int = pow(2, 18)):
        def set_label(text):
            self.label.textContent = f'{file.name}: {text}'

        logger.info(f'upload file: {file.name} {file.size} {file.type}')
        set_label('uploading...')
        try:
            from server import rpc
            await rpc.upload_init(file.name, file.size)
            offset = 0
            total_size = file.size
            self.progress.max = total_size
            while offset < total_size:
                chunk: js.Blob = file.slice(offset, offset + chunk_size)
                array_buffer = await self._read_chunk(chunk)
                # Process the chunk_text as needed
                logger.info(f'offset={offset}')
                b64str = base64.b64encode(array_buffer.to_py()).decode()
                await rpc.upload_append(file.name, b64str)
                offset += chunk_size
                self.progress.value = offset
                # percentage with two decimals
                percentage = round(offset / total_size * 100, 2)
                set_label(f'{percentage}%')
            set_label('done')
            self.progress.value = total_size
        except Exception as e:
            set_label(f'error: {e}')
            logger.exception(e)
            raise

    async def _read_chunk(self, blob: js.Blob) -> str:
        future: asyncio.Future = asyncio.Future()

        def onload(event):
            future.set_result(reader.result)

        def onerror(event):
            future.set_exception(JsException(reader.error))

        reader: js.FileReader = js.FileReader.new()
        reader.onload = create_proxy(onload)
        reader.onerror = create_proxy(onerror)
        reader.readAsArrayBuffer(blob)

        return await future



import js
import logging

import wwwpy.remote.component as wpc

from .upload_component import UploadComponent
from .rpc_issue import RpcIssueAlert

logger = logging.getLogger(__name__)


class Component1(wpc.Component, tag_name='component-1'):
    multiple_checkbox: js.HTMLInputElement = wpc.element()
    upload1: UploadComponent = wpc.element()

    def init_component(self):
        # language=html
        self.element.innerHTML = """
<div style='padding: 0.5em; margin: 0.5em; line-height: 1.5'>
        <div>Component1 in component1.py</div>
<div>The following component is defined in file upload_component.py</div>
<div>The files are uploaded in the project root 'uploads' folder. To change this behaviour see file server/rpc.py</div>
<label style='display: block'>
    <input data-name="multiple_checkbox" placeholder="input1" type="checkbox">
 Multiple files upload</label>
 
<wwwpy-quickstart-upload data-name="upload1"></wwwpy-quickstart-upload>  
 </div>               
        """
        self.multiple_checkbox.checked = self.upload1.multiple
        self.element.append(RpcIssueAlert().element)

    async def multiple_checkbox__input(self, event):
        js.console.log('handler multiple_checkbox__input event =', event)
        self.upload1.multiple = self.multiple_checkbox.checked





from __future__ import annotations

from threading import Lock
from typing import Any, Callable, List
from datetime import datetime, timedelta


class Debouncer:
    def __init__(self, window: timedelta, wakeup: Callable[['Debouncer'], None] = None,
                 time_func: Callable[[], datetime] = datetime.utcnow):
        self.window = window
        self.wakeup = wakeup
        self._time_func = time_func
        self._events: List[Any] = []
        self._last_event_time: datetime | None = None
        self._lock = Lock()

    @property
    def is_debouncing(self) -> bool:
        """True if we were in active window OR we have events yet to be emitted."""
        with self._lock:
            return self._last_event_time is not None

    def add_event(self, event: Any) -> None:
        """Add an event to the debouncer.
        If the event is the first event in the window, wakeup() will be called to signal that the waiting time is changed.
        """
        with self._lock:
            current_time = self._time_func()
            self._events.append(event)
            self._last_event_time = current_time
            if len(self._events) == 1:
                self.wakeup(self)

    def events(self) -> List[Any]:
        with self._lock:
            emission = self._time_until_next_emission()
            if emission is None or emission > timedelta(0):
                return []

            events_to_emit = self._events.copy()
            self._events.clear()
            self._last_event_time = None

            return events_to_emit

    def time_until_next_emission(self) -> timedelta:
        with self._lock:
            return self._time_until_next_emission()

    def _time_until_next_emission(self) -> timedelta:
        if not self._events or self._last_event_time is None:
            return self.window

        current_time = self._time_func()
        time_since_last_event = current_time - self._last_event_time
        time_remaining = self.window - time_since_last_event

        return time_remaining



from pathlib import Path
from typing import List

from wwwpy.common.collectionlib import ListMap
from wwwpy.common.designer.element_library import ElementDef, EventDef, Help, NamedListMap, AttributeDef
from wwwpy.common.rpc import serialization

parent = Path(__file__).parent


#
def _reorder(elements: List[ElementDef]):
    hide = {'sl-drawer', 'sl-radio', 'sl-dialog', 'sl-include', 'sl-menu-label',
            'sl-radio-button', 'sl-animation', 'sl-breadcrumb', 'sl-popup',
            'sl-breadcrumb-item', 'sl-carousel-item', 'sl-menu-item', 'sl-tree-item'}
    for e in elements.copy():
        if e.tag_name in hide:
            elements.remove(e)

    order = ['sl-button', 'sl-input', 'sl-textarea',
             'sl-checkbox', 'sl-switch', 'sl-progress-bar', 'sl-split-panel',
             'sl-textarea', 'sl-alert', 'sl-select', 'sl-dropdown']

    order.reverse()
    by_tag_name = {element.tag_name: element for element in elements}

    for o in order:
        ed = by_tag_name.get(o)
        if not ed:
            raise ValueError(f"Element {o} not found")
        elements.remove(ed)
        elements.insert(0, ed)


def _set_icon_values(elements: List[ElementDef], tag_names: List[str]):
    icon_values = (parent / 'sl_icons.txt').read_text().split('\n')
    for tag_name in tag_names:
        element = elements.get(tag_name)
        if not element:
            continue
        attribute = element.attributes.get('name')
        if not attribute:
            continue
        attribute.values = icon_values


def _insert_event(elements: List[ElementDef], tag_names, event_def: EventDef):
    for tag_name in tag_names:
        element = elements.get(tag_name)
        if not element:
            continue
        element.events.insert(0, event_def)


def _insert_attribute(elements: List[ElementDef], tag_names, attribute_name: str):
    for tag_name in tag_names:
        element: ElementDef = elements.get(tag_name)
        if not element:
            continue
        attr = AttributeDef(attribute_name)
        element.attributes.insert(0, attr)


def _shoelace_elements_def() -> List[ElementDef]:
    shoelace_json = (parent / 'shoelace.json').read_text()
    elements = serialization.from_json(shoelace_json, List[ElementDef])
    elements = ListMap(elements, key_func=lambda x: x.tag_name)
    for element in elements:
        element.gen_html = _shoelaceGenerateHtml

    _set_icon_values(elements, ['sl-icon', 'sl-icon-button'])
    _insert_event(elements, ['sl-button', 'sl-icon-button', 'sl-input',
                             'sl-textarea', 'sl-checkbox', 'sl-progress-bar'], EventDef(
        'click', Help('', 'https://developer.mozilla.org/en-US/docs/Web/API/Element/click_event')))

    _reorder(elements)
    return elements


def _shoelaceGenerateHtml(element_def: ElementDef, name: str) -> str:
    tag_name = element_def.tag_name

    def _def(placeHolder=False, add=''):
        def inner():
            pl = '' if not placeHolder else f' placeholder="{name}"'
            add1 = '' if not add else f' {add}'
            return f'<{tag_name} data-name="{name}"{pl}{add1}>{name}</{tag_name}>'

        return inner

    func = {
        'sl-button': _def(),
        'sl-input': lambda: f"""<sl-input data-name="{name}" placeholder="{name}"></sl-input>""",
        'sl-textarea': _def(placeHolder=True),
        'sl-select': lambda: f'''<sl-select data-name="{name}" label="Select a Few" value="option-2" multiple clearable max-options-visible="0">
    <sl-option value="option-1">Option 1</sl-option>
    <sl-option value="option-2">Option 2</sl-option>
    <sl-option value="option-3">Option 3</sl-option>
</sl-select>''',
        'sl-checkbox': _def(),
        'sl-radio-group': lambda: f"""<sl-radio-group data-name="{name}" label="Select an option" name="a" value="1">
    <sl-radio value="1">Option 1</sl-radio>
    <sl-radio value="2">Option 2</sl-radio>
    <sl-radio value="3">Option 3</sl-radio>
</sl-radio-group>""",
        'sl-switch': _def(),
        'sl-alert': lambda: f"""<sl-alert data-name="{name}" variant="primary" open closable>
    <sl-icon slot="icon" name="info-circle"></sl-icon>
    You can close this alert any time!
</sl-alert>""",
        'sl-badge': _def(),
        'sl-card': lambda: f'''<sl-card data-name="{name}" class="card-overview">
    <img slot="image" src="https://images.unsplash.com/photo-1559209172-0ff8f6d49ff7?ixlib=rb-1.2.1&ixid=eyJhcHBfaWQiOjEyMDd9&auto=format&fit=crop&w=500&q=80" alt="A kitten sits patiently between a terracotta pot and decorative grasses."/>
    <strong>Mittens</strong><br />
    This kitten is as cute as he is playful. Bring him home today!<br />
    <small>6 weeks old</small>
    <div slot="footer">
        <sl-button variant="primary" pill>More Info</sl-button>
        <sl-rating></sl-rating>
    </div>
</sl-card>''',
        'sl-dialog': lambda: f'''<sl-dialog data-name="{name}" label="{name}" open><sl-button slot="footer" variant="primary">Close</sl-button></sl-dialog>''',
        'sl-drawer': lambda: f'''<sl-drawer data-name="{name}" open label="Drawer" placement="end" class="drawer-placement-start">
    <span>Drop elements here.<br>To close the drawer with Python set {name}.open=False<br>In html, remove the attribute `open`</span>
</sl-drawer>''',
        'sl-dropdown': lambda: f'''<sl-dropdown data-name="{name}">
    <sl-button slot="trigger" caret>{name}</sl-button>
    <sl-menu>
        <sl-menu-item>Item 1</sl-menu-item>
        <sl-menu-item>Item 2</sl-menu-item>
        <sl-menu-item>Item 3</sl-menu-item>
    </sl-menu>
</sl-dropdown>''',
        'sl-tooltip': lambda: f'''<sl-tooltip data-name="{name}" content="{name}"><sl-button>Hover Me</sl-button></sl-tooltip>''',
        'sl-progress-bar': _def(add='value="50"'),
        'sl-image-comparer': lambda: f'''<sl-image-comparer data-name="{name}">
    <img slot="before" src="https://images.unsplash.com/photo-1517331156700-3c241d2b4d83?ixlib=rb-1.2.1&ixid=eyJhcHBfaWQiOjEyMDd9&auto=format&fit=crop&w=800&q=80&sat=-100&bri=-5" alt="Grayscale version of kittens in a basket looking around." />
    <img slot="after" src="https://images.unsplash.com/photo-1517331156700-3c241d2b4d83?ixlib=rb-1.2.1&ixid=eyJhcHBfaWQiOjEyMDd9&auto=format&fit=crop&w=800&q=80" alt="Color version of kittens in a basket looking around." />
</sl-image-comparer>''',
        'sl-avatar': lambda: f"""<sl-avatar data-name="{name}" 
    image="https://images.unsplash.com/photo-1591871937573-74dbba515c4c?ixlib=rb-1.2.1&auto=format&fit=crop&w=300&q=80"
    label="Avatar of a white and grey kitten on grey textile"></sl-avatar>""",
        'sl-skeleton': lambda: """<div style="display: block; width: 70%">
  <header style="display: flex; align-items: center; margin-bottom: 1rem;">
    <sl-skeleton style="float: left; width: 3rem; height: 3rem; margin-right: 1rem; vertical-align: middle;"></sl-skeleton>
    <sl-skeleton style="flex: 0 0 auto; width: 30%;"></sl-skeleton>
  </header>

  <sl-skeleton style="width: 95%; margin-bottom: 1rem;"></sl-skeleton>
  <sl-skeleton style="width: 80%; margin-bottom: 1rem;"></sl-skeleton>
  <sl-skeleton style="width: 60%;  margin-bottom: 1rem;"></sl-skeleton>
</div>""",
        'sl-spinner': _def(),
        'sl-icon': _def(add='name="star"'),
        'sl-icon-button': lambda: f'''<sl-icon-button data-name="{name}" name="house-gear"></sl-icon-button>''',
        'sl-split-panel': lambda: f'''<sl-split-panel data-name="{name}" position="25" style="height: 300px; --divider-width: 20px; background: var(--sl-color-neutral-50);">
    <sl-icon slot="divider" name="grip-vertical"></sl-icon>
    <div slot="start" style="display: flex; align-items: center; justify-content: center; overflow: hidden;">
        <span>Start</span>
    </div>
    <div slot="end" style="display: flex; align-items: center; justify-content: center; overflow: hidden;">
        <span>End<br>You can change orientation setting the attribute `vertical`</span>
    </div>
</sl-split-panel>''',
        'sl-tree': lambda: f"""<sl-tree data-name="{name}" selection="multiple">
    <sl-tree-item expanded><sl-icon name="folder"></sl-icon>
        Documents
        <sl-tree-item><sl-icon name="folder"></sl-icon>
            Photos
            <sl-tree-item><sl-icon name="image"></sl-icon>birds.jpg</sl-tree-item>
            <sl-tree-item><sl-icon name="image"></sl-icon>kitten.jpg</sl-tree-item>
            <sl-tree-item><sl-icon name="image"></sl-icon>puppy.jpg</sl-tree-item>
        </sl-tree-item>

        <sl-tree-item>
            <sl-icon name="folder"></sl-icon>Writing
            <sl-tree-item><sl-icon name="file"></sl-icon>draft.txt</sl-tree-item>
            <sl-tree-item><sl-icon name="file-pdf"></sl-icon>final.pdf</sl-tree-item>
            <sl-tree-item><sl-icon name="file-bar-graph"></sl-icon>sales.xls</sl-tree-item>
        </sl-tree-item>
    </sl-tree-item>
</sl-tree>""",
        'sl-progress-ring': lambda: f"""<sl-progress-ring data-name="{name}" value="35"></sl-progress-ring>""",
        'sl-carousel': lambda: f"""<sl-carousel data-name="{name}" autoplay pagination navigation loop mouse-dragging>
    <sl-carousel-item><img src="https://images.unsplash.com/photo-1464822759023-fed622ff2c3b" alt="Green mountain across body of water"></sl-carousel-item>
    <sl-carousel-item><img src="https://images.unsplash.com/photo-1501785888041-af3ef285b470" alt="three brown wooden boat on blue lake water taken at daytime"></sl-carousel-item>
    <sl-carousel-item><img src="https://images.unsplash.com/photo-1532274402911-5a369e4c4bb5" alt="brown wooden dock between lavender flower field near body of water during golden hour"></sl-carousel-item>
    <sl-carousel-item><img src="https://images.unsplash.com/photo-1523712999610-f77fbcfc3843" alt="forest heat by sunbeam"></sl-carousel-item>
    <sl-carousel-item><img src="https://images.unsplash.com/photo-1472214103451-9374bd1c798e" alt="green grass field during sunset"></sl-carousel-item>
    <sl-carousel-item><img src="https://images.unsplash.com/photo-1501785888041-af3ef285b470" alt="three brown wooden boat on blue lake water taken at daytime"></sl-carousel-item>
</sl-carousel>""",
        'sl-details': lambda: f"""<sl-details data-name="{name}" summary="Toggle Me">
  Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor incididunt ut labore et dolore magna
  aliqua. Ut enim ad minim veniam, quis nostrud exercitation ullamco laboris nisi ut aliquip ex ea commodo consequat.
</sl-details>""",
        'sl-animated-image': lambda: f"""<sl-animated-image data-name={name} src="https://shoelace.style/assets/images/tie.webp" alt="Animation of a shoe being tied"></sl-animated-image>"""

    }
    gen_html = func.get(tag_name, None)
    html = '\n' + gen_html() if gen_html else '' + ElementDef.default_gen_html(element_def, name)
    return html



from __future__ import annotations

import sys
from pathlib import Path
from typing import List

from wwwpy.common import modlib

import logging

logger = logging.getLogger(__name__)


def is_component(path: Path):
    return (path.is_file() and path.suffix == '.py' and
            path.name.startswith('component'))


def next_name(current_files: List[str]) -> tuple[int, str]:
    files = set(current_files)
    for i in range(1, sys.maxsize):
        name = f'component{i}.py'
        if name not in files:
            return i, name
    raise ValueError('Really?')


def main():
    add()


def add(folder: Path | None = None) -> Path | None:
    if folder is None:
        folder = modlib._find_package_directory('remote')
        if folder is None:
            logger.error('remote package not found')
            return None

    index, name = next_name([d.name for d in folder.iterdir() if is_component(d)])
    file = folder / name
    assert not file.exists()
    content = f'''import wwwpy.remote.component as wpc
import js

import logging

logger = logging.getLogger(__name__)

class Component{index}(wpc.Component, tag_name='component-{index}'):
    def init_component(self):
        # language=html
        self.element.innerHTML = """<span>component-{index}</span>"""
'''
    file.write_text(content)
    return file


if __name__ == '__main__':
    main()



from __future__ import annotations

from datetime import timedelta
from pathlib import Path
from typing import List, Callable, Set

from wwwpy.common import modlib, files
from wwwpy.common.filesystem import sync
from wwwpy.common.filesystem.sync import Sync
from wwwpy.common.filesystem.sync import sync_delta2
from wwwpy.remote.designer.rpc import DesignerRpc
from wwwpy.server.filesystem_sync.any_observer import logger
from wwwpy.server.filesystem_sync.watchdog_debouncer import WatchdogDebouncer
from wwwpy.websocket import WebsocketPool, PoolEvent

sync_impl: Sync = sync_delta2

import logging

logger = logging.getLogger(__name__)


def _watch_filesystem_change_for_remote(package: str, websocket_pool: WebsocketPool):
    directory = modlib._find_package_directory(package)
    if not directory:
        return

    def on_sync_events(events: List[sync.Event]):
        try:
            filt_events = _remove_blacklist(events, directory, package)
            if len(filt_events) > 0:
                payload = sync_impl.sync_source(directory, filt_events)
                for client in websocket_pool.clients:
                    remote_rpc = client.rpc(DesignerRpc)
                    remote_rpc.package_file_changed_sync(package, payload)
        except:
            # we could send a sync_init
            import traceback
            logger.error(f'on_sync_events 1 {traceback.format_exc()}')

    handler = WatchdogDebouncer(directory, timedelta(milliseconds=100), on_sync_events)
    handler.watch_directory()


def _hotreload_remote(remote_packages: list[str], websocket_pool: WebsocketPool, ):
    for package in remote_packages:
        _watch_filesystem_change_for_remote(package, websocket_pool)


def _watch_filesystem_change_for_server(package: str, callback: Callable[[str, List[sync.Event]], None]):
    directory = modlib._find_package_directory(package)
    if not directory:
        return

    def on_sync_events(events: List[sync.Event]):
        try:
            # oh, boy. When a .py file is saved it fires the first hot reload. Then, when that file is loaded
            # the python updates the __pycache__ files, firing another (unwanted) reload: the first was enough!
            filt_events = _remove_blacklist(events, directory, package)
            if len(filt_events) > 0:
                callback(package, filt_events)
        except:
            import traceback
            logger.error(f'_watch_filesystem_change_for_server {traceback.format_exc()}')

    handler = WatchdogDebouncer(directory, timedelta(milliseconds=100), on_sync_events)
    handler.watch_directory()


def _hotreload_server(hotreload_packages: list[str]):
    def on_change(package: str, events: List[sync.Event]):

        for p in hotreload_packages:
            directory = modlib._find_package_directory(p)
            if directory:
                try:
                    import wwwpy.common.reloader as reloader
                    reloader.unload_path(str(directory))
                except:
                    # we could send a sync_init
                    import traceback
                    logger.error(f'_hotreload_server {traceback.format_exc()}')

    for package in hotreload_packages:
        _watch_filesystem_change_for_server(package, on_change)


def _remove_blacklist(events: List[sync.Event], directory: Path, package: str) -> List[sync.Event]:
    def reject(e: sync.Event) -> bool:
        src_path = Path(e.src_path)
        if src_path.suffix in files.extension_blacklist:
            return True
        p = src_path.relative_to(directory)
        for part in p.parts:
            if part in files.directory_blacklist:
                return True
        return False

    result = [e for e in events if not reject(e)]
    _print_events(result, directory, package, len(events) - len(result))
    return result


def _warning_on_multiple_clients(websocket_pool: WebsocketPool):
    def pool_before_change(event: PoolEvent):
        client_count = len(websocket_pool.clients)
        if client_count > 1:
            logger.warning(f'WARNING: more than one client connected, total={client_count}')
        else:
            logger.warning(f'Connected client count: {client_count}')

    websocket_pool.on_after_change.append(pool_before_change)


def _print_events(events: List[Event], root_dir: Path, package: str, blackisted_count: int):
    def accept(e: Event) -> bool:
        bad = e.is_directory and e.event_type == 'modified'
        return not bad

    def to_str(e: Event) -> str:
        def rel(path: str) -> str:
            return str(Path(path).relative_to(root_dir))

        dest_path = rel(e.dest_path) if e.dest_path else ''
        src_path = rel(e.src_path)
        return src_path if dest_path == '' else f'{src_path} -> {dest_path}'

    summary = list(set(to_str(e) for e in events if accept(e)))
    blacklist_applied = f' [{blackisted_count} blacklisted events]' if blackisted_count > 0 else ''
    logger.info(f'Hotreload package `{package}` events: {len(events)}. Changes summary: {summary}{blacklist_applied}')



from __future__ import annotations

from enum import Enum
from typing import NamedTuple, Protocol, List

from wwwpy.common.rpc.serializer import RpcRequest


class WebsocketRoute(NamedTuple):
    path: str
    on_connect: Callable[[WebsocketEndpointIO], None]


class Change(Enum):
    add = 'add'
    remove = 'remove'


class PoolEvent(NamedTuple):
    change: Change
    endpoint: WebsocketEndpoint
    pool: WebsocketPool

    @property
    def add(self) -> bool:
        return self.change == Change.add

    @property
    def remove(self) -> bool:
        return self.change == Change.remove


class PoolChangeCallback(Protocol):
    def __call__(self, change: PoolEvent) -> None: ...


class WebsocketPool:

    def __init__(self, route: str):
        self.clients: list[WebsocketEndpoint] = []
        self.http_route = WebsocketRoute(route, self._on_connect)
        self.on_before_change: List[PoolChangeCallback] = []
        self.on_after_change: List[PoolChangeCallback] = []

    def _notify_change(self, change: PoolEvent, listeners: List[PoolChangeCallback]) -> None:
        for callback in listeners:
            callback(change)

    def _on_connect(self, endpoint: WebsocketEndpointIO) -> None:

        add = PoolEvent(Change.add, endpoint, self)
        self._notify_change(add, self.on_before_change)

        def handle_remove(msg: str | bytes | None):
            if msg is None:
                remove = PoolEvent(Change.remove, endpoint, self)
                self._notify_change(remove, self.on_before_change)
                self.clients.remove(endpoint)
                self._notify_change(remove, self.on_after_change)

        endpoint.listeners.append(handle_remove)
        self.clients.append(endpoint)
        self._notify_change(add, self.on_after_change)


class ListenerProtocol(Protocol):
    def __call__(self, message: str | bytes | None) -> None: ...


class SendEndpoint:
    def send(self, message: str | bytes | None) -> None: ...


class DispatchEndpoint(Protocol):
    def dispatch(self, module: str, func_name: str, *args) -> None: ...


from typing import TypeVar, Callable

T = TypeVar('T')


class WebsocketEndpoint(SendEndpoint):
    listeners: list[ListenerProtocol]

    # part to be called by user code to send a outgoing message
    def send(self, message: str | bytes | None) -> None: ...

    def rpc(self, factory: Callable[..., T]) -> T:
        instance = factory(self)
        return instance


class WebsocketEndpointIO(WebsocketEndpoint):
    def __init__(self, send: ListenerProtocol):
        """The send argument is called by the IO implementation: it will deliver outgoing messages"""
        self._send = send
        self.listeners: list[ListenerProtocol] = []

    # part to be called by user code to send a outgoing message
    def send(self, message: str | bytes | None) -> None:
        self._send(message)

    # parte to be used by IO implementation to be called to notify incoming messages
    def on_message(self, message: str | bytes | None) -> None:
        for listener in self.listeners:
            listener(message)

    def dispatch(self, module: str, func_name: str, *args) -> None:
        j = RpcRequest.build_request(module, func_name, *args).json()
        self.send(j)



from __future__ import annotations

import asyncio
from dataclasses import dataclass
from enum import Enum
from typing import Optional, List

import js
import wwwpy.remote.component as wpc
from pyodide.ffi import create_proxy
from wwwpy.common import state, property_monitor
from wwwpy.common.designer import element_library, html_locator, code_strings
from wwwpy.common.designer.element_editor import ElementEditor, EventEditor
from wwwpy.common.designer.element_path import ElementPath
from wwwpy.remote import dict_to_js
from wwwpy.remote.designer.helpers import _element_path_lbl, _rpc_save, _log_event, _help_button

from .button_tab import ButtonTab, Tab
from .searchable_combobox2 import SearchableComboBox, Option


# write enum class with [events, attributes] and use it in the button click event
class PropertyEditorMode(str, Enum):
    palette = 'palette'
    events = 'events'
    attributes = 'attributes'


@dataclass
class PropertyEditorState:
    mode: PropertyEditorMode = PropertyEditorMode.events


class PropertyEditor(wpc.Component, tag_name='wwwpy-property-editor'):
    row_container: js.HTMLElement = wpc.element()
    message1div: js.HTMLElement = wpc.element()
    _tabs: ButtonTab = wpc.element()
    _selected_element_path: Optional[ElementPath] = None
    state: PropertyEditorState = PropertyEditorState()

    @property
    def selected_element_path(self) -> Optional[ElementPath]:
        return self._selected_element_path

    @selected_element_path.setter
    def selected_element_path(self, value: Optional[ElementPath]):
        self._selected_element_path = value
        self._render()

    def init_component(self):
        self._state_manager = state.State(state.JsStorage(), 'wwwpy.toolbar.property-editor.state')
        self.state = self._state_manager.restore(PropertyEditorState).instance_or_default()
        property_monitor.monitor_changes(self.state, lambda *a: self._state_manager.save(self.state))

        # language=html
        self.element.innerHTML = """
<style>
        .wwwpy-property-editor {
            display: grid;
            grid-template-columns: 2fr 3fr;
            box-sizing: content-box;
        }
        .wwwpy-property-editor-row {
            display: contents;
        }
        .wwwpy-property-editor-row > :first-child {
            display: flex;
            align-items: center;
            padding-left: 5px;
        }
        .wwwpy-property-input {
            width: 100%;
            padding: 3px;
            box-sizing: border-box;
            /* border: 1px solid ; */                     
        }
    </style>
<div data-name="message1div" style='color: white'>&nbsp</div>
<wwwpy-button-tab data-name="_tabs"></wwwpy-button-tab>
<div class="wwwpy-property-editor" data-name='row_container'>    
    <div class='wwwpy-property-editor-row' style='font-weight: bold'><div >Event</div><div>Value</div></div>
</div>

        """

        def set_state_render(mode: PropertyEditorMode):
            self.state.mode = mode
            self._render()

        self._tabs.tabs = [
            Tab('palette', on_selected=lambda tab: set_state_render(PropertyEditorMode.palette)),
            Tab('events', on_selected=lambda tab: set_state_render(PropertyEditorMode.events)),
            Tab('attributes', on_selected=lambda tab: set_state_render(PropertyEditorMode.attributes)),
        ]

        for lbl in ['data-name', 'name', 'type', 'value', 'form']:
            row1 = PropertyEditorRowAttribute2()
            row1.element.classList.add('wwwpy-property-editor-row')
            self.row_container.appendChild(row1.element)
            row1.label.innerHTML = lbl
            row1.value.placeholder = 'Double click creates handler'
            if lbl == 'type':
                row1.value.value = 'text'

    def add_row(self, row: wpc.Component):
        row.element.classList.add('wwwpy-property-editor-row')
        self.row_container.appendChild(row.element)

    def _render(self):
        ep = self._selected_element_path
        self.message1div.innerHTML = '' if ep is None else f'Selection: {_element_path_lbl(ep)}'

        self.row_container.innerHTML = ''
        if not ep:
            return
        lib = element_library.element_library()
        element_def = lib.by_tag_name(ep.tag_name)
        if not element_def:
            self.message1div.innerHTML += '<br>No metadata found for editing.'
            return
        self.message1div.innerHTML = f'Selection: {_element_path_lbl(ep)} {_help_button(element_def)}'
        html = code_strings.html_from(ep.class_module, ep.class_name)
        if not html:
            self.message1div.innerHTML += '<br>No HTML found for editing.'
            return

        cst_node = html_locator.locate_node(html, ep.path)
        if cst_node is None:
            self.message1div.innerHTML += '<br>No element found in the HTML.'
            return

        self.row_container.innerHTML = ''
        if self.state.mode == PropertyEditorMode.attributes:
            self._render_attribute_editor(element_def, ep)
        elif self.state.mode == PropertyEditorMode.events:
            self._render_event_editor(element_def, ep)

    def _set_title(self, lbl, value):
        pe_title = PropertyEditorTitleRow()
        pe_title.label.innerHTML = lbl
        pe_title.value.innerHTML = value
        self.add_row(pe_title)

    def _render_attribute_editor(self, element_def, ep):
        self._set_title('Attribute', 'Value')
        element_editor = ElementEditor(ep, element_def)
        for attr_editor in element_editor.attributes:
            attr_def = attr_editor.definition
            row1 = PropertyEditorRowAttribute() if not attr_def.boolean else PropertyEditorRowBoolAttribute()
            self.add_row(row1)
            row1.label.innerHTML = attr_def.name
            if not attr_def.boolean:
                options: List[Option] = [Option(value) for value in attr_def.values]
                focus_search = len(options) > 0
                for option in options:
                    if option.text == '':
                        option.label = 'Set to empty string'
                        option.italic = True

                if attr_editor.exists:
                    if not attr_def.mandatory:
                        remove = Option('remove attribute')
                        remove.actions.set_input_value = False
                        remove.italic = True

                        def remove_selected(ae=attr_editor):
                            ae.remove()
                            self._save(element_editor)

                        remove.on_selected = remove_selected  # lambda ae=attr_editor: ae.remove()
                        options.insert(0, remove)
                click_for = 'Click for options...'
                if attr_def.boolean:
                    row1.value.placeholder = 'attribute present' if attr_editor.exists else click_for
                else:
                    pre_placeholder = 'Set to empty string. ' if attr_editor.exists and attr_editor.value == '' else ''
                    row1.value.placeholder = pre_placeholder + (click_for if len(options) > 0 else '')
                row1.value.option_popup.options = options
                row1.value.text_value = '' if attr_editor.value is None else attr_editor.value
                row1.value.option_popup.search_placeholder = 'Search options...'
                row1.value.focus_search_on_popup = focus_search

                def attr_changed(event, ae=attr_editor, row=row1):
                    js.console.log(f'attr_changed {ae.definition.name} {row.value.text_value}')
                    ae.value = row.value.text_value
                    self._save(element_editor)

                row1.value.element.addEventListener('wp-change', create_proxy(attr_changed))
            else:
                if attr_editor.exists:
                    row1.value.checked = True
                else:
                    row1.value.checked = False

                def attr_changed(event, ae=attr_editor, row=row1):
                    if row.value.checked:
                        ae.value = None  # beware, this is creating the attribute with no value
                    else:
                        ae.remove()
                    self._save(element_editor)

                row1.value.addEventListener('change', create_proxy(attr_changed))

    def _save(self, element_editor):
        source = element_editor.current_python_source()

        async def start_save():
            await _rpc_save(self._selected_element_path, source)

        asyncio.create_task(start_save())

    def _render_event_editor(self, element_def, ep):
        self._set_title('Event', 'Value')
        element_editor = ElementEditor(ep, element_def)
        for event_editor in element_editor.events:
            row1 = PropertyEditorRowAttribute2()
            self.add_row(row1)
            row1.label.innerHTML = event_editor.definition.name
            row1.value.placeholder = '' if event_editor.handled else 'Double click creates handler'
            row1.value.readOnly = True
            row1.value.value = event_editor.method.name if event_editor.handled else ''

            def dblclick(ev: EventEditor = event_editor):
                js.console.log(f'dblclick on {ev.definition.name}')
                handled = ev.handled
                ev.do_action()
                source = element_editor.current_python_source()

                async def start_save():
                    if not handled:
                        await _rpc_save(ep, source)
                    await _log_event(element_editor, ev)

                asyncio.create_task(start_save())

            row1.double_click_handler = lambda ev=event_editor: dblclick(ev)


class PropertyEditorRow(wpc.Component, tag_name='wwwpy-property-editor-row'):
    label: js.HTMLElement = wpc.element()
    value: js.HTMLElement = wpc.element()

    def root_element(self):
        return self.shadow

    def init_component(self):
        self.shadow = self.element.attachShadow(dict_to_js({'mode': 'open'}))
        # language=html
        self.shadow.innerHTML = """
        <slot name="label" data-name="label"></slot>
        <slot name="value" data-name="value"></slot>
            """


class PropertyEditorRowAttribute2(wpc.Component):
    label: js.HTMLElement = wpc.element()
    value: js.HTMLInputElement = wpc.element()
    double_click_handler = None

    def init_component(self):
        # language=html
        self.element.innerHTML = """
        <div data-name="label">uff</div><input data-name='value' type="text" class="wwwpy-property-input">
            """

    def value__dblclick(self, event):
        if self.double_click_handler:
            self.double_click_handler()


class PropertyEditorRowAttribute(wpc.Component):
    label: js.HTMLElement = wpc.element()
    value: SearchableComboBox = wpc.element()
    double_click_handler = None

    def init_component(self):
        # language=html
        self.element.innerHTML = """
        <div data-name="label">uff</div>
        <wwwpy-searchable-combobox2 data-name='value' class="wwwpy-property-input"></wwwpy-searchable-combobox2>
            """

    def value__dblclick(self, event):
        if self.double_click_handler:
            self.double_click_handler()


class PropertyEditorRowBoolAttribute(wpc.Component):
    label: js.HTMLElement = wpc.element()
    value: js.HTMLInputElement = wpc.element()

    def init_component(self):
        # language=html
        self.element.innerHTML = """
        <div data-name="label">uff</div>
        <input type="checkbox" data-name="value" style="transform: scale(1.4)">
            """


class PropertyEditorTitleRow(wpc.Component):
    label: js.HTMLElement = wpc.element()
    value: js.HTMLElement = wpc.element()

    def init_component(self):
        # language=html
        self.element.innerHTML = """
        <div style='font-weight: bold' data-name="label">Event</div><div style='font-weight: bold' data-name='value'>Value</div></div>        
            """



import asyncio

from js import console
from wwwpy.common.designer import log_emit
from wwwpy.server.designer import rpc

_log_buffer = []


async def _process_buffer():
    """This techique avoids the issue of wrong log order"""
    loc = _log_buffer.copy()
    _log_buffer.clear()

    for msg in loc:
        await rpc.server_console_log(msg)


def redirect_logging():
    def emit(msg: str):
        console.log(msg)
        _log_buffer.append(msg)
        asyncio.create_task(_process_buffer())

    log_emit.add_once(emit)
    log_emit.warning_to_log()
